// controller.h

/******************************************************************************
 *
 *  MiXViews - an X window system based sound & data editor/processor
 *
 *  Copyright (c) 1993, 1994 Regents of the University of California
 *
 *  Author:     Douglas Scott
 *  Date:       December 13, 1994
 *
 *  Permission to use, copy and modify this software and its documentation
 *  for research and/or educational purposes and without fee is hereby granted,
 *  provided that the above copyright notice appear in all copies and that
 *  both that copyright notice and this permission notice appear in
 *  supporting documentation. The author reserves the right to distribute this
 *  software and its documentation.  The University of California and the author
 *  make no representations about the suitability of this software for any 
 *  purpose, and in no event shall University of California be liable for any
 *  damage, loss of data, or profits resulting from its use.
 *  It is provided "as is" without express or implied warranty.
 *
 ******************************************************************************/


// The Controller class is one of the three central classes in MiXViews.  It is
// named such because of its role in the Model-View-Controller paradigm on
// which MiXViews is based.  It coordinates and routes messages between various
// other graphic and non-graphic components.  It contains pointers to an Editor
// subclass instance, a DataView subclass instance, and a DataWindow instance.
// It controls the creation and destruction of each display window and all of
// its associated parts.  Each is connected via a doubly-linked circular list,
// each instance being a link in the list.  Operations that pass data between
// windows do so via methods invoked upon the Controller via the DataEditor 
// subclass member, and the entire list can be searched from anywhere on itself.

#ifndef CONTROLLER_H
#ifdef __GNUG__
#pragma interface
#endif
#define CONTROLLER_H

#include <InterViews/resource.h>
#include "localdefs.h"
#include "range.h"
#include "MyString.h"

class DataFile;
class Data;
class DataWindow;
class ProgressDialog;
class DataView;
class ViewInfo;
class World;
class Event;
class DataEditor;
class Request;
class ProgressAction;
class EditStatusDelegate;

class Controller : public Resource {
private:
	static Controller*	_sHead;
	static int			_sNumControllers;
	static int			_sIsQuitting;
	Controller*			_prev;
	Controller*			_next;
	static int			_sAvailableSources;
	enum ViewType { AsChannels, AsFrames };
	static ViewInfo&	DefaultViewInfo;
private:
	void addToList();
	void removeFromList();
	boolean closeChain();
public:
	static Controller* create(DataFile *);
	Controller(Data *, const char *name = nil, ViewInfo &info=DefaultViewInfo);

	void display(World *);
	const char *fileName() const;
	const char* windowName () const;
	Data* model() const;
	DataEditor* editor() const { return _editor; }

	void busy(boolean);

	// methods used (called by) Graphs	
	void setEditRegion(const Range &region, int chan);
	void setInsertPoint(int point, int chan);

	// methods used by views
	void setCommandState(unsigned int, bool set);

	// methods used by editor
	World *world();	
	boolean handleEvent(Event &);
	boolean handleKeyEvent(Event &);
	int handleRequest(Request &);
	boolean keyCommand(unsigned long);
	void notifyOfSelection(bool);
	Data* findSelection();
	void showInsertPoint(int, const Range &, boolean scroll = false);
	void showEditRegion(const Range &, const Range &, boolean scroll = false);
	void unselectView();
	void resetScaleTimes();
	void alert(const char* m1, const char* m2=nil,
		const char* m3=nil, const char* ="  ok  ");
	boolean confirm(const char* m1, const char* m2=nil,
		const char* m3=nil, Response r=Yes, const char* ="confirm",
		const char* ="cancel");
	Response choice(const char* m1, const char* m2=nil,
		const char* m3=nil, Response r=Yes, const char* =" yes ",
		const char* ="  no  ", const char* ="cancel");

	// methods used by Modifiers
	ProgressAction* createProgressAction(const char *message);
	int modifyProgress(double);		// For modifier progress

	// methods called via menus or keyboard
	void newViewOfSelection();
	void viewAsFrames();
	void viewAsChannels();
	void zoomToSelection();
	void zoomToFull();
	boolean close();
	void showVersion();
	void showManual();
	void showHomePage();
	void quit();
	void changeName();
	void setProgramOptions();
protected:
	Controller(const char *);
	virtual ~Controller();
	Data* getSelection(long &timestamp);
	void externalSourceNotify(bool);
	void setFileName(const char *);
	void viewAs(ViewType type);
 	// converter methods
	ProgressAction* getConverterProgressAction() { return _converterProgressAction; }
private:
	void earlyInit();
	boolean initialize(ViewInfo &info=DefaultViewInfo);
	DataView* newView(ViewInfo &);
	void createAndAttachView(ViewInfo &info=DefaultViewInfo);
	void detachView();
	int playbackProgress(double);	// For playback/record progress action
private:
	String				_filename;
	DataWindow *		_window;
	DataView *			_view;
	DataEditor *		_editor;
	ProgressAction *	_converterProgressAction;
	ProgressDialog *	_progressDialog;
	EditStatusDelegate*	_editDelegate;
	boolean				_converting;
	friend class DataEditor;
	friend class SoundEditor;
	friend class FFTEditor;
	friend class LPCEditor;
	friend class PCHEditor;
	friend class EnvelopeEditor;
	friend class PvocEditor;
	friend class ProcessFunction;
	friend class FileSaver;
};

inline const char *
Controller::fileName() const { return _filename; }

#endif
